//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================

/** 

\mainpage Enigma User Manual

\section toc Table of Contents

- \ref levels "Creating Enigma Levels"

*/


/** 

\page levels Creating Enigma Levels

\section lintro Introduction

Enigma does not currently include a graphical level editor.  Instead,
landscapes are described by small programs that are loaded and
interpreted by the game engine.

To create new levels you have to know a few basics about Lua, the
programming language used internally by Enigma.  Fortunately, Lua is a
very simple language, and only a subset of it is required for all but
the most complicated levels.

The bulk of this chapter describes the available game objects, their
attributes, and the programming interface used to put these objects in
a landscape.  This manual will not teach you how to create *new* kinds
of objects or modify existing ones.  If that's what you're interested
in, your best bet is to (a) read the source code, and (b) join the
Enigma development mailing list.

This file is more of a reference manual than a tutorial.  If you are
new to creating Enigma levels, you may prefer to read the \ref
loverview Overview in the next section, study some of the existing
levels, and then come back to this text for more information as
needed.

- \ref loverview: An example level
- \ref lstones

*/

/** \page loverview Quick Overview

Here is a very simple level description that can also serve as a
starting-point for new landscapes.  (In fact, this is the first level
in Enigma, so you can try it out right away.)

\verbatim
 1   create_world(20, 13) 
 2   draw_border("st-brownie") 
 3   fill_floor("fl-hay", 0,0, level_width,level_height)
 4
 5   set_stone("st-fart", level_width-1,0, {name="fart"})
 6   set_stone("st-timer", 0,0, {action="trigger", target="fart",
 7             interval=10})
 8   enigma.AddConstantForce(0,6)
 9
10   oxyd(3,3)
11   oxyd(level_width-4,level_height-4) 
12   oxyd(level_width-4, 3) 
13   oxyd(3,level_height-4) 
14   oxyd_shuffle() 
15 
16   set_actor("ac-blackball", 10,6.5, {player=0})
\endverbatim

Let's now turn to a line-by-line analysis of this ``program'':

\verbatim
 1   create_world(20, 13) 
 2   draw_border("st-brownie") 
 3   fill_floor("fl-hay", 0,0, level_width,level_height)
\endverbatim

The first line creates a new world that is 20 blocks wide and 13
blocks high.  Each ``block'' in the world can be accessed with using a
pair of coordinates: The upper left corner has coordinates (0,0), the
lower right one has coordinates (19,12).  Each block contains a floor
tile, an (optional) item, and an (optional) stone.

A frame of stones is drawn around the landscape just created with the
``draw_border'' command.  The argument, \c "st-brownie", is the name
of a stone.  By convention, all stones have "st-" prefixed to their
name, similarly all item names begin with "it-" and all floor names
with "fl-".

The "fill_floor" command in line 3 likewise fills the complete floor
with tiles of type "fl-hay".  As additional arguments, this function
takes the upper left corner and the width and height of the rectangle
to be filled.

\verbatim
 5   set_stone("st-fart", level_width-1,0, {name="fart"})
 6   set_stone("st-timer", 0,0, {action="trigger", target="fart",
 7             interval=10})
\endverbatim

Lines 5--7 demonstrate how to put individual stones in the world.  The
``set_stone'' command takes a stone name, the desired coordinates, and
an (optional) list of ``attributes'' as arguments.  Note the use of
curly braces `{', `}' to enclose the attribute list.

Attributes are the key to customizing the behaviour of objects in a
landscape.  Here, we first give a "name" to the first stone we create.
It's a fart stone that has the unpleasant habit of "blowing off" when
triggered.  Triggering this fart stone is done by the timer stone we
create in line 6--7.  This stone performs a predefined action at
regular intervals.  In this case we want to send a "trigger" message
every ten seconds to the object named "fart".

\verbatim
 8   enigma.AddConstantForce(0,6)
\endverbatim

This command adds a constant force field to the landscape.  The
horizontal component of the force is 0, while the vertical is 6, so
all movable objects in the landscape (in this case we have only one
black marble) are pulled towards the bottom of the screen.

\verbatim
10   oxyd(3,3)
11   oxyd(level_width-4,level_height-4)
12   oxyd(level_width-4, 3)
13   oxyd(3,level_height-4)
14   oxyd_shuffle()
\endverbatim

These commands place a couple of oxyd stones in the level.  The \c
oxyd command internally uses ``set_stone("st-oxyd", x,y, ...)'' to
create the stones, but it additionally assigns sensible values to some
of the oxyd stones' attributes (most notably the color).  The command
on line 14 permutes the colors on the oxyd stones currently in the
landscape.

\verbatim
16   set_actor("ac-blackball", 10,6.5, {player=0})
\endverbatim

This final line creates the black marble controlled by the player.
Objects that can move around freely are called \e actors in Enigma
Unlike stones and items, actors are \e not restricted to integer
coordinates, as can be seen in this example.  The final argument is,
as usual, a list of attributes.  For the time being, this should
always be "{player=0}".

[The "player" attribute will become more important in future releases
when one player can control multiple actors, or when there are two (or
even more) players.]
*/

/** \page lstones  Available Stones

\ref st-coinslot
\ref st-fakeoxyd
\ref st-laser
\ref st-oxyd
\ref st-pmirror
\ref st-puzzle
\ref st-timer
\ref st-3mirror
*/

#include "config.h"
#include "enigma.hh"
#include "display.hh"
#include "world.hh"
#include "objects.hh"
#include "player.hh"
#include "lua.hh"
#include "sound.hh"
#include "system.hh"
#include "gui.hh"
#include "video.hh"

#include "px/px.hh"

#include "SDL.h"
#include "getopt.h"

#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <string>
#include <cassert>
#include <iostream>
#include <algorithm>

using namespace std;
using namespace px;
using namespace enigma;


namespace {
    void optionsmenu();
    void gamemenu();

    void flush_events()
    {
        SDL_Event e;
        while (SDL_PollEvent(&e)) ;
    }
}

void 
GridPos::move(Direction dir) 
{ 
    switch(dir) { 
    case NORTH: y--; break; case SOUTH: y++; break;
    case EAST: x++; break; case WEST: x--; break;
    case NODIR: break;
    }
}

GridPos 
enigma::move(GridPos p, Direction dir) 
{
    GridPos tmp = p;
    tmp.move(dir);
    return tmp;
}

bool 
enigma::operator!=(GridPos a, GridPos b)
{
    return (a.x!=b.x || a.y!=b.y);
}

Direction 
enigma::reverse(Direction d)
{
    Direction rdir[] = { EAST, NORTH, WEST, SOUTH };
    return d==NODIR ? NODIR : rdir[d];
}

string
enigma::to_suffix(Direction d)
{
    char *sfx[] = { "", "-w", "-s", "-e", "-n" };
    return sfx[d+1];
}


//======================================================================
// MENUS
//======================================================================
namespace 
{
    using gui::TextButton;

    class GameMenu : public gui::Menu, gui::ActionListener {
    public:
        GameMenu() 
            : resume(new TextButton("Resume Game", this)),
              restart(new TextButton("Restart Level", this)),
//              options(new TextButton("Options", this)),
              main(new TextButton("Main Menu", this))
        {
            add(resume,     Rect(0,0,150,40)); 
            add(restart,    Rect(0,50,150,40)); 
//            add(options,    Rect(0,100,150,40)); 
            add(main,       Rect(0,100,150,40));
            center(video::GetScreen());
        }
    private:

        void on_action(gui::Widget *w) {
            if (w == resume)
                Menu::quit();
            else if (w == main) {
                enigma::QuitGame();
                Menu::quit();
            }
            else if (w == restart)
            {
                enigma::RestartLevel();
                Menu::quit();
            }
            else
            {
            }
        }
        gui::Widget *resume, *restart, *options, *main;
    };

    class ImageAlloc {
    public:
        Surface *acquire(const std::string &name) {
            return px::LoadImage(name.c_str());
        }
        void release(Surface *s) { delete s; }
    };

    typedef cache::Cache<Surface*, ImageAlloc> ImageCache;

}


//======================================================================
// MAIN PROGRAM
//======================================================================

namespace
{
    struct LevelInfo 
    {
        LevelInfo(string fn, string n, string a)
            : filename(fn), name(n), author(a)
        {
            hidden=false;
            playable=true;
            completed=false;
        }

        string filename;
        string name;
        string author;
        bool hidden;
        bool playable;
        bool completed;
    };

    vector<LevelInfo> levels;

    vector<string>  args;       // list of command line arguments
    bool quit_mainmenu;

    px::Screen *get_screen() { return video::GetScreen(); } 
}


//----------------------------------------
// LevelMenu implementation
//----------------------------------------
namespace
{
    class LevelMenu {
    public:
        LevelMenu();
        int manage(Screen *scr);
    private:
        void paint(Screen *scr);
        void set_selected(int idx);
        void set_first(int first);

        void scroll_up(int lines)
        {
            for (; lines; --lines) {
                if (ifirst+width*height > (int)levels.size())
                    break;
                ifirst += width;
                if (iselected < ifirst)
                    iselected += width;
                redraw=true;
            }
        }

        void scroll_down(int lines)
        {
            for (; lines; --lines) {
                if (ifirst < width)
                    break;
                ifirst -= width;
                if (iselected >= ifirst+width*height)
                    iselected -= width;
                redraw=true;
            }
        }

        // Variables.
        ImageCache cache;
        int ifirst;             // index of upper left level
        int iselected;          // index of selected level
        int width, height;
        Font &smallfnt;
        vector<Rect> areas;
        bool redraw;
        bool show_filenames_p;
    };
}

LevelMenu::LevelMenu()
    : ifirst(0), iselected(0),
      width(3), height(3),
      smallfnt(enigma::GetFont("levelmenu")),
      areas(),
      redraw(true),
      show_filenames_p(false)
{}

void
LevelMenu::set_first(int first)
{
    if (first != ifirst)
    {
        ifirst = first;
        redraw=true;
    }
}

void 
LevelMenu::set_selected(int idx)
{
    if (idx < 0)
        idx = 0;
    if (unsigned(idx) >= levels.size())
        idx = levels.size()-1;

    int first = ifirst;
    while (idx < first)
        first -= width;
    while (idx >= first + width*height)
        first += width;

    if (first != ifirst || idx != iselected)
    {
        redraw = true;
        ifirst = first;
        iselected = idx;
    }
}

void
LevelMenu::paint(Screen *screen)
{
    GC gc(screen);
    video::HideMouse();

    areas.clear();

    set_color(gc, 0,0,0);
    Surface *bg = &enigma::GetImage("menu_bg");
    blit(gc, 0,0, bg);
//    box(gc, screen->size());


    const int imgw = 160;
    const int imgh = 104;
    const int hgap = 20, vgap=40;

    const int x0 = (screen->width() - width*(imgw+hgap)+hgap)/2;
    const int y0 = 30;

    unsigned i=ifirst;
    for (int y=0; y<height; y++)
        for (int x=0; x<width; x++, i++) {
            if (i >= levels.size())
                goto done_painting;

            int xpos = x0 + x*(imgw + hgap);
            int ypos = y0 + y*(imgh + vgap);
            areas.push_back(Rect(xpos-hgap/2, ypos-hgap/2, 
                                 imgw+hgap, imgh+hgap+vgap/2));

            string fname = string("levels/") + levels[i].filename + ".png";
            if (Surface *img = cache.get(enigma::FindDataFile(fname)))
                screen->blit(xpos, ypos, img);

            if (levels[i].completed) {
                screen->blit(xpos, ypos, &enigma::GetImage("completed"));
            }
            
            char str[10];
            sprintf(str, "#%d  ",i+1);//(%d)   ", i);

            string tstr(str);
            if (show_filenames_p)
                tstr += levels[i].filename;
            else
                tstr += levels[i].name;
            auto_ptr<Surface> title(smallfnt.render(tstr));
            screen->blit(xpos+(imgw-title->width())/2,
                         ypos+imgh+1,
                         title.get());
        }
  done_painting:

    set_color(gc, 255,0,0);
    Rect a=areas[iselected-ifirst];
    frame(gc, a);
    frame(gc, smaller(a,1));
    frame(gc, smaller(a,2));

    video::ShowMouse();
    screen->update_all();
}

int
LevelMenu::manage(Screen *scr)
{
    redraw = true;
    SDL_Event e;
    while (true) {
        scr->flush_updates();
        if (redraw) {
            paint(scr);
            redraw=false;
        }
        SDL_Delay(10);

        if (SDL_PollEvent(&e)) 
        {
            if (e.type == SDL_MOUSEBUTTONDOWN) {
                switch (e.button.button) {
                case 1:
                    for (unsigned i=0; i<areas.size(); ++i)
                        if (areas[i].contains(e.button.x, e.button.y))
                        {
                            sound::PlaySound("menuok");
                            return ifirst+i;
                        }
                    break;
                case 4: scroll_down(1); break;
                case 5: scroll_up(1); break;
                }
            }
	    else if (e.type == SDL_KEYDOWN) {
                int newsel = iselected;
                int newfirst = ifirst;
                
                switch (e.key.keysym.sym) {
                case SDLK_t: 
                    show_filenames_p=!show_filenames_p; 
                    redraw=true;
                    break;
                case SDLK_LEFT: newsel--; break;
                case SDLK_RIGHT: newsel++; break;
                case SDLK_DOWN: newsel += width; break;
                case SDLK_UP: newsel -= width; break;
                case SDLK_PAGEDOWN: 
                    newfirst += width*height; 
                    newsel   += width*height;
                    break;
                case SDLK_PAGEUP:
                    newfirst -= width*height; 
                    newsel   -= width*height;
                    break;
                case SDLK_HOME: newsel = 0; break;
                case SDLK_END: newsel = levels.size()-1; break;
                case SDLK_RETURN:
                    return iselected;
                case SDLK_ESCAPE:
                    return -1;
                default:
                    break;
                }
                if (newfirst!=ifirst && newfirst>=0 && newfirst<(int)levels.size())
                {
                    sound::PlaySound("menumove");
                    ifirst = newfirst;
                    iselected = Max(Min(newsel,(int)levels.size()-1), 0);
                    redraw=true;
                }
                else if (newsel!=iselected && newsel>=0 && newsel<(int)levels.size())
                {
                    sound::PlaySound("menumove");
                    while (newsel < ifirst)
                        ifirst -= width;
                    while (newsel >= ifirst+width*height)
                        ifirst += width;
                    iselected = newsel;
                    redraw = true;
                }
            }
        }

    }
    cache.clear();
}


//----------------------------------------------------------------------
// Level list
//----------------------------------------------------------------------

void 
enigma::ClearLevelList()
{
    levels.clear();
}

void 
enigma::AddLevel(const std::string &filename,
                 const std::string &name)
{
    levels.push_back(LevelInfo(filename, name, ""));
}

//----------------------------------------------------------------------
// GAME
//----------------------------------------------------------------------
namespace
{
    class Game {
    public:
        Game();
        void run(int ilevel);
        void quit() { change_state(ABORT); }
        void next_level() { change_state(NEXTLEVEL); }
        void restart_level() { change_state(RESTARTLEVEL); }
    private:
        void load_level(int ilevel);

        // Private variables.
        
        enum State 
        { 
            /* The game is currently running. */
            INGAME,

            /* This level has been completed, proceed to the next
               one. */
            NEXTLEVEL,

            /* Player's marble is dead. Wait a little, then restart
               the level. */
            PLAYERDEAD,

            /* Restart the current level (without necessarily
               reloading it! */
            RESTARTLEVEL,
 
            /* Reload the current level (this automatically resets
               every object in the landscape). */
            RELOADLEVEL,

            /* The level info screen before entering a new level. */
            LEVELINFO,  

            /* Leave the game immediately. */
            ABORT
        } state;

        enum InputState {
            I_NORMAL, I_MESSAGE, I_COMMAND
        } input_state;

        unsigned icurrent_level;
        Uint32 last_tick_time;
        double level_finished_dtime;
        double actor_dead_dtime;

        // Private methods.
        void handle_events();
        void on_keydown(SDL_Event &e);
        void on_mousebutton(SDL_Event &e);
        void change_state(State newstate);
        void tick(double dtime);
        void show_menu();
    };
}

Game::Game()
    : state(INGAME), 
      input_state(I_NORMAL),
      icurrent_level(0),
      last_tick_time(0)
{
}

void 
Game::change_state(State newstate)
{
    if (state == newstate) return;
    state = newstate;
    switch (state) {
    case NEXTLEVEL:
        level_finished_dtime=0;
        display::ShowText("Level finished!", display::TEXT_STATIC);
        break;
    case PLAYERDEAD:
        actor_dead_dtime=0;
//        display::ShowText("You lost", display::TEXT_STATIC);
        break;
    default: break;
    }
}

void
Game::tick(double dtime)
{
    switch (state) {
    case RESTARTLEVEL:
        load_level(icurrent_level);
        change_state(INGAME);
        break;

    case PLAYERDEAD:
        actor_dead_dtime += dtime;
        if (actor_dead_dtime <= 0.5) {
            handle_events();
            world::Tick(dtime);
            display::Tick(dtime);
            display::Redraw(get_screen());
        } else {
            change_state(RELOADLEVEL);
        }
        break;

    case RELOADLEVEL:
        load_level(icurrent_level);
        change_state(INGAME);
        break;

    case LEVELINFO:
        // TODO: show level information (name, author, etc.)
        change_state(INGAME);
        break;

    case NEXTLEVEL:
        level_finished_dtime += dtime;
        if (level_finished_dtime <= 1.5) {
            handle_events();
            world::Tick(dtime);
            display::Tick(dtime);
            display::Redraw(get_screen());
        } else {
            if (1+icurrent_level == levels.size()) 
                change_state(ABORT);
            else {
                levels[icurrent_level].completed = true;
                load_level(1+icurrent_level);
                change_state(LEVELINFO);
            }
        }
        break;

    case INGAME:

        if (player::AllActorsDead()) {
            change_state(PLAYERDEAD);
        }

        handle_events();
        world::Tick(dtime);
        display::Tick(dtime);
        {
//            Uint32 t = SDL_GetTicks();
            display::Redraw(get_screen());
//            printf ("%g\n", 1000.0/(SDL_GetTicks()-t));
        }
        break;

    default:
        break;
    }
}

void 
Game::load_level(int ilevel) 
{
    Fade(video::FADEOUT);
    
    icurrent_level=ilevel;
    world::Load(levels[icurrent_level].filename);

    display::Redraw(get_screen(), false);
    Fade(video::FADEIN);
    flush_events();
    last_tick_time = SDL_GetTicks();
}

void
Game::run(int ilevel)
{
    player::NewGame(1);         // one player

//    sound::PlayMusic("/sound/Emilie.xm");
    sound::FadeoutMusic();
    sound::StopMusic();
    load_level(ilevel);

    double dtime=0;

    state = INGAME;
    while (state != ABORT) 
    {
        last_tick_time=SDL_GetTicks();

        tick(dtime);

        int sleeptime = 10 - (SDL_GetTicks()-last_tick_time);
        if (sleeptime > 0)
            SDL_Delay(sleeptime);

        dtime=(SDL_GetTicks()-last_tick_time)/1000.0;
	if (dtime > 500.0) /* Time has done something strange, perhaps
			      run backwards */
            dtime = 0.0;
	else if (dtime > 0.5)
            dtime = 0.5;
    }
}

void
Game::handle_events()
{
    SDL_Event e;
    while (SDL_PollEvent(&e)) 
    {
        switch (e.type) {
        case SDL_KEYDOWN:
            on_keydown(e);
            break;
        case SDL_MOUSEMOTION: 
            {
                V3 force;
                force[0] = e.motion.xrel;
                force[1] = e.motion.yrel;
                double mspeed = lua::get_num("MouseSpeed");
                world::SetMouseForce(force*mspeed);
            }
            break;
        case SDL_MOUSEBUTTONDOWN:
        case SDL_MOUSEBUTTONUP:
            on_mousebutton(e);
            break;
        case SDL_ACTIVEEVENT:
            if( e.active.gain == 0)
                show_menu();
            break;
        case SDL_QUIT:
            change_state(ABORT);
            break;
        }
    }
}

void
Game::on_mousebutton(SDL_Event &e)
{
    if (e.button.state == SDL_PRESSED)
    {
        if (e.button.button == 1) {
            // left mousebutton -> activate first item in inventory
            player::ActivateItem();
            player::InhibitPickup(true);
        }
        else if (e.button.button == 3) {
            // right mousebutton -> rotate inventory
            player::RotateInventory();
            player::InhibitPickup(true);
        }
    } 
    else {
        // mouse button released -> allow picking up items
        int b = SDL_GetMouseState(0, 0);
        if (!(b & SDL_BUTTON(1)) &&  !(b & SDL_BUTTON(3))) {
            player::InhibitPickup(false);
        }
    }
}

void
Game::on_keydown(SDL_Event &e)
{
    switch (e.key.keysym.sym) {
    case SDLK_ESCAPE:
        show_menu();
        break;
    case SDLK_F10:
        {
            string fname = levels[icurrent_level].filename + ".bmp"; 
            video::Screenshot(fname.c_str());
        }
        break;
    case SDLK_F3:
        player::Suicide();
        break;
    case SDLK_x:
        if (e.key.keysym.mod & KMOD_ALT) {
            change_state(ABORT);
            quit_mainmenu = true;
        }
        break;
    default:
        break;
    }

    if (lua::get_num("WizardMode") > 0.0) {
        switch (e.key.keysym.sym) {
        case SDLK_n:
            enigma::FinishLevel();
            break;
        case SDLK_l:
            load_level(icurrent_level);
            break;
        case SDLK_t:
            TintRect(get_screen(), get_screen()->size(), 
                     0,0,0, 200);
            get_screen()->update_all();
            get_screen()->flush_updates();
            break;
        case SDLK_g:
            if (e.key.keysym.mod & KMOD_ALT) {
                display::ReloadModels();
                load_level(icurrent_level);
            }
            break;
        case SDLK_1: ToggleFlag(display::SHOW_FLOOR); break;
        case SDLK_2: ToggleFlag(display::SHOW_ITEMS); break;
        case SDLK_3: ToggleFlag(display::SHOW_SHADES); break;
        case SDLK_4: ToggleFlag(display::SHOW_STONES); break;
        case SDLK_5: ToggleFlag(display::SHOW_SPRITES); break;
        default:
            break;
        }
    }
}

void 
Game::show_menu()
{
    Screen *scr = get_screen();
//     TintRect(scr, scr->size(), 0,0,0, 150);
//     scr->update_all();

    video::ShowMouse();
    GameMenu m;
    m.manage(scr);
    video::HideMouse();
    last_tick_time = SDL_GetTicks();
    if (state != ABORT)
        display::RedrawAll(get_screen());
}

namespace
{
    Game game_inst;
}

static void
game()
{
    if (levels.empty())
        lua::Dofile(enigma::FindDataFile("levels/index.lua"));

    if (levels.empty()) {
        cerr << "Error: No levels defined\n";
        return;
    }

    int ilevel=-1;
    {
        LevelMenu lm;
        ilevel = lm.manage(get_screen());
    }
    if (ilevel == -1)
        return;                 // return to main menu

    try {
        video::HideMouse();
        game_inst.run(ilevel);
        video::ShowMouse();
    } 
    catch (lua::Error)
    {
        fprintf(stderr, "Error in Lua script\n");
        world::Shutdown();
    }
}

void enigma::FinishLevel() {
    game_inst.next_level();
}

void enigma::RestartLevel() {
    game_inst.restart_level();
}

void enigma::QuitGame() {
    game_inst.quit();
}

//----------------------------------------------------------------------
// Main menu
//----------------------------------------------------------------------
namespace
{
    class MainMenu : public gui::Menu, public gui::ActionListener {
    public:
        MainMenu() 
            : newgame(new gui::TextButton("New Game", this)), 
              options(new gui::TextButton("Options", this)), 
              quit(new gui::TextButton("Quit", this))
        {
            int x0=(640-150)/2;
            int y0=200;
            add(newgame, Rect(x0,y0,150,40));
            add(options, Rect(x0,y0+50,150,40));
            add(quit, Rect(x0,y0+100,150,40));
        }
    private:
        // Menu interface.
        void draw_background(px::GC &gc) 
        {
            video::SetCaption("Enigma - Main Menu");
            sound::PlayMusic("sound/menu.s3m");

            blit(gc, 0,0,&enigma::GetImage("menu_bg"));

            Surface& logo(enigma::GetImage("enigma_logo3"));
            int x0=(640-logo.width())/2;
            int y0=40;
            blit(gc, x0, y0, &logo);
        }

        // ActionListener interface.
        void on_action(gui::Widget *w) {
            if (w == newgame) {
                game();
                invalidate_all();
            } else if (w == options) {
                optionsmenu();
                invalidate_all();
            } else if (w == quit) {
                Menu::quit();
            }
        }

        // Variables.
        gui::Widget *newgame, *options, *quit;
    };

    void mainmenu()
    {
        MainMenu m;

        video::SetMouseCursor(enigma::LoadImage("it-magicwand"), 4, 4);
        video::ShowMouse();

        try {
            m.manage(get_screen());
        } catch (lua::Error) {
            cerr << "Returned to main menu...\n";
        }
    }
}

namespace
{
    class FullscreenButton : public gui::TextButton, gui::ActionListener {
    public:
        FullscreenButton() : TextButton("", this) {update();}
    private:
        void update() {
            if (lua::get_num("FullScreen") > 0)
                set_text("Fullscreen");
            else
                set_text("Window");
        }
        // ActionListener interface.
        void on_action(Widget *w) {
            video::ToggleFullscreen();
            update();
        }
    };

    class AlphaShadowButton : public gui::TextButton, gui::ActionListener {
    public:
        AlphaShadowButton() : TextButton("", this) {update();}
    private:
        void update() {
            if (lua::get_num("AlphaShadows") > 0)
                set_text("Blended Shadows");
            else
                set_text("Stippled Shadows");
        }
        // ActionListener interface.
        void on_action(Widget *w) {
//            video::ToggleFullscreen();
            update();
        }
    };


    class OptionsMenu : public gui::Menu, public gui::ActionListener {
    public:
        OptionsMenu() : quit(new TextButton("Back", this))
        {
            add(new FullscreenButton, Rect(0,0,150,40));
//            add(new AlphaShadowButton, Rect(0,50,150,40));
            add(quit, Rect(0,50,150,40));
        }
    private:

        // ActionListener interface.
        void on_action(gui::Widget *w)
        {
            if (w == quit)
                Menu::quit();
        }

        // Menu interface.
        void draw_background(px::GC &gc) 
        {
            video::SetCaption("Enigma - Options Menu");
            blit(gc, 0,0,&enigma::GetImage("menu_bg"));
        }
        
        gui::Widget *quit;
    };


    void optionsmenu()
    {
        OptionsMenu m;
        m.center(get_screen());
        m.manage(get_screen());
    }

}

namespace
{
    class DataPathList {
    public:
        DataPathList(const string &pathspec=DEFAULT_DATA_PATH) 
        { set_path(pathspec); }
        
        void set_path(const string &pathspec)
        {
            path=pathspec;
            datapaths.clear();
            split_copy(pathspec, ':', back_inserter(datapaths));

            for_each(datapaths.begin(), datapaths.end(), &sysdep::expand_path);
        }
        const string &get_path() const { return path; }

        std::string find_file(const std::string &filename) const {
            for (unsigned i=0; i<datapaths.size(); ++i) {
                string complete_name = datapaths[i] + sysdep::path_separator + filename;
                if (sysdep::FileExists(complete_name))
                    return complete_name;
            }
            throw px::XFileNotFound();
        }
    private:
        string path;
        vector<string> datapaths;
    };

    DataPathList datapaths;
}

string enigma::GetDataPath() {
    return datapaths.get_path();
}

void enigma::SetDataPath(const string &p) {
    datapaths.set_path(p);
}

string enigma::FindDataFile(const string &filename) {
    try {
        return datapaths.find_file(filename);
    } catch (px::XFileNotFound x) {
        fprintf(stderr, "File not found: %s\n", filename.c_str());
        return filename;
    }
}

//----------------------------------------------------------------------
// Startup
//----------------------------------------------------------------------

namespace
{
    void load_config_file()
    {
        // First try a configuration file in the current directory.
        string fname = "enigma_conf.lua";
        try {
            lua::Dofile(enigma::FindDataFile(fname));
        } catch (lua::Error) {
            cerr << "No configuration file found.\n";
        }
    }

    void save_config_file()
    {
        string rcfile = sysdep::expand_path("~/.enigmarc");
    }

    void usage() {
        cout << "Available command line options for Enigma:\n\n"
             << "    --nosound      disable music and sound\n"
             << "    --nomusic      disable music\n"
             << "    --help -h      show this help\n"
             << "    --version      print the executable's version number\n"
             << endl;
    }
}

static void 
init()
{
    lua::Init();

    bool nosound_flag = false;
    bool nomusic_flag = false;
    bool show_help = false;
    bool show_version = false;
    for (unsigned i=0; i < ::args.size(); ++i) 
    {
        string& arg = ::args[i];

        if (arg == "--help" || arg == "-h")
            show_help = true;
	else if (arg == "--nosound")
	    nosound_flag = true;
        else if (arg == "--nomusic")
            nomusic_flag = true;
        else if (arg == "--version")
            show_version = true;
        else if (arg == "--wizard")
            lua::set_num("WizardMode", 1.0);
        else
            show_help = true;   // unknown argument
    }
    if (show_help || show_version) {
        cout << "Enigma v" << VERSION << endl;
        if (show_help) usage();
        exit(0);
    }

    // load preferences
    load_config_file();

    int sdl_flags = SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE;
    if (!nosound_flag)
	sdl_flags |= SDL_INIT_AUDIO;

    if (SDL_Init(sdl_flags) < 0) 
    {
        cerr << "Couldn't init SDL: " << SDL_GetError() << endl;
        exit(1);
    }

    if (nosound_flag)
        sound::DisableSound();
    else if (nomusic_flag)
        sound::DisableMusic();

    video::Init();
    display::Init();
    sound::Init();
    world::Init();

    // load level index
    lua::Dofile(enigma::FindDataFile("init.lua"));
}

static void 
shutdown()
{
    video::Shutdown();
    display::Shutdown();
    world::Shutdown();
    save_config_file();
}


namespace
{
    class FontAlloc {
    public:
        Font *acquire(const std::string &name) {
            return px::LoadBitmapFont(
                enigma::FindDataFile(string("fonts/") + name + ".png"),
                enigma::FindDataFile(string("fonts/") + name + ".bmf"));
        }
        void release(Font *f) { delete f; }
    };

    cache::Cache<Font *, FontAlloc> font_cache;
    cache::Cache<Surface*, ImageAlloc> image_cache;
}

px::Font *
enigma::LoadFont(const std::string &name)
{
    string png = string("fonts/") + name + ".png";
    string bmf = string("fonts/") + name + ".bmf";
    return px::LoadBitmapFont(FindDataFile(png).c_str(),
                              FindDataFile(bmf).c_str());
}

px::Font &
enigma::GetFont(const string &name)
{
    Font *f = font_cache.get(name);
    assert(f);
    return *f;
}

px::Surface *
enigma::LoadImage(const std::string &name)
{
    string filename = FindDataFile(string("gfx/") + name + ".png");
    return px::LoadImage(filename.c_str());
}

px::Surface &
enigma::GetImage(const std::string &name)
{
    string filename = FindDataFile(string("gfx/") + name + ".png");
    Surface *s = image_cache.get(filename);
    assert(s);
    return *s;
}

int
main(int argc, char** argv)
{
    try {
        copy(argv+1, argv+argc, back_inserter(::args));
        init();
        mainmenu();
        shutdown();
    }
    catch (px::XGeneric& e) {
        cerr << "main: caught px exception: " << e.get_string() << endl;
        shutdown();
        return 1;
    }
    catch (std::exception& e) {
        cerr << "main: caught exception: " << e.what() << endl;
        shutdown();
        return 1;
    }
    return 0;
}
