//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#ifndef GUI_HH_INCLUDED
#define GUI_HH_INCLUDED

#include "px/pxfwd.hh"
#include "px/geom.hh"
#include "SDL.h"
#include <vector>
#include <string>

//======================================================================
// EVENTS AND EVENT HANDLERS
//======================================================================
namespace gui
{
    class Widget;

    class ActionListener {
    public:
        virtual ~ActionListener() {}
        virtual void on_action(Widget *w) = 0;
    };
}

//======================================================================
// GUI WIDGETS
//======================================================================
namespace gui
{
    class Menu;

    class Widget {
    public:
        virtual ~Widget() {}

        virtual bool has_changed() const { return false; }
        virtual void draw(px::GC &gc) = 0;
        virtual void activate() {}
        virtual void deactivate() {}

        virtual void on_event(const SDL_Event &e) {}

        void set_pos(int x, int y) {area.x = x; area.y = y;}
        void set_size(int w, int h) {area.w = w; area.h = h;}

        px::Rect get_area() const { return area; }
        void set_area(const px::Rect &r) { area = r; }
        int get_x() const { return area.x; }
        int get_y() const { return area.y; }
        int get_w() const { return area.w; }
        int get_h() const { return area.h; }
        Menu * get_menu() const { return menu; }
    protected:
        Widget() : area(), menu(0) {}

        void redraw();
    private:

        friend class Menu;
        void set_menu(Menu *m) { menu=m; }

        px::Rect area;
        Menu *menu;
    };


    class Button : public Widget {
    protected:
        Button() : highlighted(false) {}

        void draw(px::GC &gc);
    protected:
        // Widget interface.
        void activate() { highlighted=true; redraw(); }
        void deactivate() { highlighted=false; redraw(); }
    private:

        bool highlighted;
    };

//----------------------------------------
// TextButton.
//----------------------------------------
    class TextButton : public Button {
    public:
        TextButton(const std::string &t, ActionListener *al=0);
        void set_listener(ActionListener *al) {
            listener = al; 
        }
        void set_text(const std::string &t);
    private:
        // Widget interface.
        void draw(px::GC &gc);
        void on_event(const SDL_Event &e);
        void deactivate() { pressed=false; redraw(); Button::deactivate(); }

        // Variables.
        static px::Font *menufont, *menufont_pressed;

        std::string text;
        bool pressed;
        ActionListener *listener;
    };

//----------------------------------------
// ImageButton.
//----------------------------------------
    class ImageButton : public Button {
    public:
        ImageButton(const std::string &fname_) : fname(fname_) 
        {}
    private:
        // Widget interface.
        void draw(px::GC &gc) {
            Button::draw(gc);
        }
        std::string fname;
    };

    class Menu {
    public:
        Menu() : widgets(), active_widget(0), quit_menu(false) {}
        virtual ~Menu() {}

        void manage(px::Screen *scr);
        void add(Widget *w) { widgets.push_back(w); w->set_menu(this); }
        void add(Widget *w, px::Rect r)
        {
            w->set_area(r);
            add(w);
        }
        void center(px::Screen *scr);


        void invalidate_area(const px::Rect &r);
        void invalidate_all();

        void draw_all(px::Screen *scr);
        void quit() { quit_menu=true; }
    private:
        void handle_event(const SDL_Event &e);
        Widget *find_widget(int x, int y);
        void redraw_area(const px::Rect &r);
        void track_active_widget( int x, int y );
        
        // Menu interface.
        virtual void draw_background(px::GC &gc) {}
        virtual bool on_event(const SDL_Event &e) { return false; }
        
        // Variables.
        std::vector<Widget*> widgets;
        Widget *active_widget;
        bool quit_menu;
        px::RectList dirtyrects;
        px::Screen *screen;
    };
}
#endif
