//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================

#include "items.hh"
#include "laser.hh"
#include "object_mixins.hh"
#include "display.hh"
#include "actors.hh"
#include "sound.hh"
#include <string>
#include <vector>
#include <algorithm>

using namespace items;
using namespace world;
using std::string;

//======================================================================
// ITEMS
//======================================================================

/** \page litems  Available Items

\ref it-document
\ref it-blackbomb
\ref it-shogundot
*/

using world::Item;

Item::Item(const string & name) 
    : GridObject(name)
{}

void 
Item::on_creation()
{
    set_model(get_kind());
    if (get_kind() != "it-laserbeam")
        laser::MaybeRecalcLight(get_pos());
}

void 
Item::on_removal()
{
    enigma::GridLoc l(enigma::GRID_ITEMS, get_pos());
    display::KillModel(l);
    laser::MaybeRecalcLight(get_pos());
}

display::Model*
Item::set_model(const string & mname)
{
    enigma::GridLoc l(enigma::GRID_ITEMS, get_pos());
    return display::SetModel(l, mname);
}

string 
Item::get_inventory_model() 
{
    return get_kind();
}

bool 
Item::actor_hit(Actor *actor)
{
    const double ITEM_RADIUS = 0.3;
    px::V3 item_center(get_pos().x + 0.5,
                       get_pos().y + 0.5,
                       0);
    double dist = length(actor->get_pos()-item_center);
    return (dist < ITEM_RADIUS);
}

//----------------------------------------
// Various simple items
//----------------------------------------
#define DEF_ITEM(classname, kindname) \
    class classname : public CloneItem<classname> { \
    public: classname() : Item(kindname) {} }

namespace
{
    DEF_ITEM(Hammer, "it-hammer");
    DEF_ITEM(MagicWand, "it-magicwand");
    DEF_ITEM(Umbrella, "it-umbrella");
    DEF_ITEM(Floppy, "it-floppy");
    
    class ExtraLife : public CloneItem<ExtraLife> {
    public:
        ExtraLife() : Item("it-extralife") {}
    private:
        std::string get_inventory_model() { return string("inv-blackball"); }
    };

    class Key : public CloneItem<Key> {
    public:
    	Key() : Item("it-key") { set_attrib("keycode", 0.0); }
    };
    class Key_a : public CloneItem<Key_a> {
    public:
    	Key_a() : Item("it-key_a") { set_attrib("keycode", 1.0); }
    };
    class Key_b : public CloneItem<Key_b> {
    public:
    	Key_b() : Item("it-key_b") { set_attrib("keycode", 2.0); }
    };
    class Key_c : public CloneItem<Key_c> {
    public:
    	Key_c() : Item("it-key_c") { set_attrib("keycode", 3.0); }
    };
}


//----------------------------------------
// Coins
//
// :value    1,2,4: how many time-units this coin buys
//----------------------------------------
namespace
{
    class Coin : public CloneItem<Coin> {
    public:
        Coin() : Item("it-coin") { set_attrib("value", 1.0); }
    private:
        void on_creation() {set_model();}
        void set_model() { Item::set_model(get_inventory_model()); }
        string get_inventory_model() 
        {
            int value=int_attrib("value");
            switch(value) {
            case 1: return "it-coin1";
            case 2: return "it-coin2";
            default:return "it-coin4";
            }
        }
    };
}

//----------------------------------------
// Hollows & Hills
//----------------------------------------

/** \page it-hills Hills and Hollows

*/
namespace
{
    class HillHollow : public StaticItem {
    protected:
        HillHollow(double radius, double forcefactor) {
            r=radius; forcefac=forcefactor;
        }

        V3 vec_to_center (V3 v)
        {
            GridPos p = get_pos();
            return v - V3(p.x+0.5, p.y+0.5,0);
        }
        double get_radius() const { return r; }
    private:
        double r, forcefac;

        px::V3 get_force(Actor *a) 
        {
            V3 v = vec_to_center(a->get_pos());
            if (length(v) <= r) 
                return forcefac*v;
            else
                return px::V3(); 
        }
    };

    class Hill : public CloneItem<Hill>, public HillHollow {
    public:
        Hill() : Item("it-hill"), HillHollow(0.5, 70) {}
    };
    class TinyHill : public CloneItem<TinyHill>, public HillHollow {
    public:
        TinyHill() : Item("it-tinyhill"), HillHollow(0.25, 70) {}
    };

    class TinyHollow : public CloneItem<TinyHollow>, public HillHollow {
    public:
        TinyHollow() : Item("it-tinyhollow"), HillHollow(0.25, -70) {}
    };

    /*
     * Hollows are special in that they can end the current level
     * if they have each a small white marble inside them.
     */
    class Hollow : public HillHollow {
    public:
        Hollow() : Item("it-hollow"), HillHollow(0.5, -70), whiteball(0) {}
        
        // Object interface.
        Object *clone() {
            Hollow *h = new Hollow;
            instances.push_back(h);
            return h;
        }
        void dispose() {
            instances.erase(find(instances.begin(), instances.end(), this));
            delete this;
        }

        // Item interface.
        bool actor_hit(Actor *a)
        {
            if (whiteball==0 && a->get_kind() == "ac-whiteball-small"
                && near_center_p(a))
            {
                whiteball = a;
                enter_time = SDL_GetTicks();
            }
            else if (whiteball==a)
            {
                if (!near_center_p(a))
                    whiteball=0;
                else
                    check_if_level_finished();
            }
            return false;
        }

        // Functions.
        bool near_center_p (Actor *a)
        {
            return (length(vec_to_center(a->get_pos())) < get_radius()/2);
        }

        /* If every hollow in the level (a) contains a small white
           ball and (b) this ball has been inside the hollow for at
           least MINTIME milliseconds, finish the level. */
        void check_if_level_finished() {
            const unsigned MINTIME=1000; 
            for (unsigned i=0; i<instances.size(); ++i) {
                Hollow *h=instances[i];
                if (!h->whiteball || SDL_GetTicks() - h->enter_time < MINTIME)
                    return;
            }
            enigma::FinishLevel();
        }
        
        // Variables.
        Actor *whiteball;
        Uint32 enter_time;
        static std::vector<Hollow *> instances;
    };
}

std::vector<Hollow*> Hollow::instances;

//----------------------------------------
// Springs
//----------------------------------------
namespace
{
    class Spring1 : public CloneItem<Spring1> {
    public:
        Spring1() : Item("it-spring1") {}
    private:
        ItemAction activate(Actor *a, GridPos)
        {
            SendMessage(a, "jump");
            return ITEM_KEEP;
        }
    };

    class Spring2 : public CloneItem<Spring2> {
    public:
        Spring2() : Item("it-spring2") {}
    private:
        ItemAction activate(Actor *a, GridPos)
        {
            SendMessage(a, "jump");
            return ITEM_KILL;
        }
    };
}

//----------------------------------------
// Explosion
//----------------------------------------
namespace
{
    class Explosion : public CloneItem<Explosion> {
    public:
        enum Strength { 
            WEAK,               // Explode but do nothing else.
            MEDIUM,             // Explode and leave a hole in the ground.
            STRONG              // Explode and shatter the floor.
        };
        Explosion(Strength st=WEAK) : Item("it-explosion"), strength(st) {}

    private:
        void on_creation() {
            set_model("expl")->set_callback(px::make_cb(this, &Explosion::animcb));
        }
        bool actor_hit(Actor *actor) {
            SendMessage(actor, "shatter");
            return false;
        }
        void animcb();

        // Variables.
        Strength strength;
    };
}

void 
Explosion::animcb() 
{
    switch (strength) {
    case WEAK:
        world::KillItem(get_pos());
        break;
    case MEDIUM:
        SetItem(get_pos(), MakeItem("it-hollow"));
        break;
    case STRONG:
        SendMessage(GetFloor(get_pos()), "explode");
        world::KillItem(get_pos());
        break;
    }
}


//----------------------------------------
// Document.
//----------------------------------------

/** \page it-document Document Item

This item looks like a piece of paper and contains text messages
that can be displayed by activating the item.
 
\subsection documenta Attributes
 
\b text:     The message to be displayed.
 
\subsection documente Example
\verbatim
set_item("it-document", 1,1, {text="Hello World!"})
\endverbatim

or, equivalently,
\verbatim
document(1,1, "Hello World")
\endverbatim
\image html it-document.png
*/

namespace
{
    class Document : public CloneItem<Document> {
    public:
        Document() : Item("it-document") {
            set_attrib("text", "");
        }

        ItemAction activate(Actor *a, GridPos)
        {
            display::ShowText(to_string(*get_attrib("text")));
            return ITEM_KILL;	       // remove from inventory
        }
        void message(const string &msg, const Value &val) {
            if (msg == "ignite")
                SetItem(get_pos(), new Explosion(Explosion::WEAK));
        }
    };
}

//----------------------------------------
// Dynamite.
//----------------------------------------
namespace
{
    class Dynamite : public CloneItem<Dynamite> {
    public:
        Dynamite() : Item("it-dynamite"), state(IDLE) {}
    private:
        enum State { IDLE, BURNING };
        State state;

        void change_state(State newstate) {
            if (newstate==BURNING && state==IDLE) {
                state = BURNING;
                set_model("it-dynamite-burning")
                    ->set_callback(px::make_cb(this, &Dynamite::explode));
            }
        }

        void explode();

        void message(const string &msg, const Value &val) {
            if (msg == "ignite")
                change_state(BURNING);
            else if (msg == "explode")
                explode();
        }

        bool on_laserhit(Direction) {
            change_state(BURNING);
            return false;           // block light
        }

        void on_drop(Actor *a) {change_state(BURNING);}

        bool actor_hit(Actor *a) {
            // don't pick up burning dynamite
            return (state == IDLE); 
        }
    };
}

void 
Dynamite::explode() 
{
    GridPos p = get_pos();

    // ignite nearby objects
    SendMessage(GetItem(move(p,enigma::NORTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::SOUTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::EAST)), "ignite");
    SendMessage(GetItem(move(p,enigma::WEST)), "ignite");

    sound::PlaySound("explosion2");
    SetItem(p, new Explosion(Explosion::MEDIUM));
}

//----------------------------------------
// BlackBomb.
//----------------------------------------

/** \page it-blackbomb Black Bomb

When black bombs explode, they destroy the floor tile underneath them.
\image html it-blackbomb.png
*/

namespace
{
    class BlackBomb : public CloneItem<BlackBomb>, public StaticItem {
    public:
        BlackBomb() : Item("it-blackbomb"), state(IDLE) {}

    private:
        enum State { IDLE, BURNING } state;

        void explode();
        void change_state(State newstate) {
            if (newstate==BURNING && state==IDLE) {
                state = BURNING;
                set_model("it-blackbomb-burning")
                    ->set_callback(px::make_cb(this, &BlackBomb::explode));
            }
        }

        void message(const string &msg, const Value &val) {
            if (msg == "ignite")
                change_state(BURNING);
            else if (msg == "explode")
                explode();
        }

        bool on_laserhit(Direction) {
            change_state(BURNING);
            return false;           // block light
        }
    };
}

void 
BlackBomb::explode() 
{
    GridPos p = get_pos();
    SendMessage(GetItem(move(p,enigma::NORTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::SOUTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::EAST)), "ignite");
    SendMessage(GetItem(move(p,enigma::WEST)), "ignite");

    sound::PlaySound("explosion1");
    SetItem(p, new Explosion(Explosion::STRONG));
}

//----------------------------------------
// Trigger.
//----------------------------------------
namespace
{
    class Trigger : public CloneItem<Trigger>, public StaticItem {
    public:
        Trigger() : Item("it-trigger"), objectcnt(0), have_stone(false) {
            set_attrib("invisible", 0.0);
        }
    private:
	bool have_stone;
        int objectcnt;          // The number of objects on the trigger

        void on_creation() {set_model();}

        void set_model() {
            if (int_attrib("invisible"))
                Item::set_model("invisible");
            else if (objectcnt > 0)
                Item::set_model("it-trigger1");
            else
                Item::set_model("it-trigger");
        }

        void add_object() {
            if (++objectcnt == 1) {
                set_model();
                world::PerformAction(this, true);
                sound::PlaySound("it-triggerdown");
            }
        }
        void remove_object() {
            if (--objectcnt == 0) {
                set_model();
                world::PerformAction(this, false);
                sound::PlaySound("it-triggerup");
            }
        }

        void actor_enter(Actor *) {
            add_object();
        }
        void actor_leave(Actor *) {
            remove_object();
        }

        void stone_change(Stone *st) {
            if (st == 0 && have_stone){
		have_stone=false;
                remove_object();
	    }
            
	    if (st != 0 && !have_stone){
		have_stone=true;
                add_object();
	    }
        }
    };
}

//----------------------------------------
// ShogunDot
//----------------------------------------

/** \page it-shogun Shogun Dot

\subsection shogundota Attributes
- \b size:            1..3  (smallest..largest)
- \b target,\b action:   as usual
*/
namespace
{
    class ShogunDot : public CloneItem<ShogunDot>, public StaticItem {
    public:
        ShogunDot();
    private:
        void set_size(int s) { set_attrib("size", s); }
        int get_size() const { return int_attrib("size"); }

        void on_creation();
        void message(const string &str, const Value &v);
        void stone_change(Stone *st);

        // Variables.
        bool activated;
    };
}

ShogunDot::ShogunDot() 
    : Item("it-shogun"), activated(false) 
{
    set_size(1);
}

void
ShogunDot::on_creation() 
{
    char *models[] = {"it-shogun1", "it-shogun2", "it-shogun3"};

    int size=get_size();
    assert(1<=size && size<=3);
    Item::set_model(models[size-1]);
}

void 
ShogunDot::stone_change(Stone *st) 
{
    if (activated && st == 0) {
        activated = false;
        world::PerformAction(this, false);
    } 
}

void 
ShogunDot::message(const string &str, const Value &v) 
{
    int size=get_size();
    if (activated) {
        if (str=="noshogun") {
            activated=false;
            world::PerformAction(this, false);
        }
    }
    else if ((size==1 && str == "shogun1") ||
            (size==2 && str == "shogun2") ||
            (size==3 && str == "shogun3"))
    {
        activated=true;
        world::PerformAction(this, true);
    }
}



void items::Init()
{
    using world::Register;

    Register(new BlackBomb);
    Register(new Coin);
    Register(new Document);
    Register(new Dynamite);
    Register(new ExtraLife);
    Register(new Floppy);
    Register(new Hammer);
    Register(new Hill);
    Register(new Hollow);
    Register(new Key);
    Register(new Key_a);
    Register(new Key_b);
    Register(new Key_c);
    Register(new MagicWand);
    Register(new ShogunDot);
    Register(new Spring1);
    Register(new Spring2);
    Register(new TinyHill);
    Register(new TinyHollow);
    Register(new Trigger);
    Register(new Umbrella);

}
