//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#include "laser.hh"
#include "sound.hh"
#include "actors.hh"
#include <algorithm>
#include <cassert>

using namespace world::laser;
using namespace world;

//======================================================================
// TYPES
//======================================================================
namespace
{
//----------------------------------------
// LaserBeam
//----------------------------------------
    class LaserBeam : public Item, public LaserEmitter {
    public:
        static void emit_from(GridPos p, Direction d);
        static void kill_all();
	
	// LaserEmitter interface
        DirectionBits emission_directions() const { return directions; }
    private:
        LaserBeam(Direction dir) : Item("it-laserbeam") {
            directions = to_bits(dir);
        }

        // Item interface.
        bool on_laserhit(Direction dir);
        void on_creation();
        void set_model();
        bool actor_hit(Actor *actor);

        Object *clone() {
            // new LaserBeams may only created inside `emit_from'.
            assert(0); 
            return 0; 
        }
        void dispose();

        // Variables
        DirectionBits directions;
        static vector<LaserBeam*> instances;
    };

//----------------------------------------
// Laser stones
//----------------------------------------

/** \page st-laser Laser Stone

These stones emit a laser beam in a specified direction when
activated.  They are the only objects in the game that can act as
primary light sources (mirrors can also emit light but they require
an incoming beam).

\subsection lasera Attributes

- \b on:      1 if laser in active, 0 if not
- \b dir:     the Direction in which light is emitted 
            (NORTH, EAST, SOUTH, WEST)

\subsection laserm Messages

- \b on, \b off, \b onoff: as usual

\subsection lasersa See also

\ref st-pmirror, \ref st-3mirror

*/
    class LaserStone : public LaserEmitter, public OnOffStone {
    public:
        LaserStone() : Stone("st-laser") {
            set_attrib("dir", Value(EAST));
        }
        static void reemit_all();

    private:
        // LaserEmitter interface
        DirectionBits emission_directions() const {
            return to_bits(get_dir());
        }

        // OnOffStone interface.
        void notify_onoff(bool on);

        // Private methods.
        void emit_light();
        Direction get_dir() const {return Direction(int_attrib("dir"));}

        // Stone interface.
        void on_creation();
        void set_model();
        Object *clone();
        void dispose();

        // Variables
        static vector<LaserStone*> instances;
    };
}

//======================================================================
// IMPLEMENTATIONS
//======================================================================

vector<PhotoCell*> PhotoCell::instances;

void
PhotoCell::notify_start() 
{
    for(unsigned i=0; i<instances.size(); ++i)
        instances[i]->on_recalc_start();
}

void
PhotoCell::notify_finish() 
{
    for(unsigned i=0; i<instances.size(); ++i)
        instances[i]->on_recalc_finish();
}

void 
PhotoCell::photo_activate() 
{
    vector<PhotoCell*>::iterator i;
    i = std::find(instances.begin(), instances.end(), this);
    if (i != instances.end())
        assert(0 || "Photocell activated twice\n");
    else
        instances.push_back(this); 
}

void 
PhotoCell::photo_deactivate() 
{
    vector<PhotoCell*>::iterator i;
    i = std::find(instances.begin(), instances.end(), this);
    if (i != instances.end())
        instances.erase(i);
}

//----------------------------------------
// PhotoStone
//----------------------------------------
void 
PhotoStone::check_state() 
{
    GridPos p = get_pos();
    bool illu = (LightFrom(p, NORTH) || LightFrom(p, EAST) 
                 || LightFrom(p, WEST) || LightFrom(p, SOUTH));
    if (illu != illuminated) {
        if (illu)
            notify_laseron();
        else
            notify_laseroff();
        illuminated = illu;
    }
}


//----------------------------------------
// LaserBeam
// 
// The implementation of laser beams is a little tricky because, in
// spite of being implemented as Items, lasers aren't localized and
// changes to any part of the beam may affect beam items elsewhere.  A
// `change' may be anything from moving a stone in or out of the beam,
// rotating or moving one of the mirrors, to making a stone in the
// beam transparent.
//
// Here are a couple of facts about laser beams in Enigma:
//
// - Laser beams are static. Once calculated they do not change until
//   they are completely recalculated
//   
// - LaserBeam::emit_from() is the only way to emit laser beams. A new
//   beam will propagate automatically and stops only if it comes
//   across a stone or or item that returns `false' from
//   on_laserhit().
//
// - `on_laserhit()' is called for objects in the beam *whenever*
//   the beam is recalculated.  For objects that need to be notified
//   when the laser goes on or off, use the `PhotoStone'
//   mixin.
//
// More bits of information can be found in the description of laser
// stones and mirrors.
//----------------------------------------

vector<LaserBeam*> LaserBeam::instances;

void 
LaserBeam::set_model() 
{
    string modelname = "it-laser";
    if (directions & (EASTBIT | WESTBIT))
        modelname += "h";
    if (directions & (NORTHBIT | SOUTHBIT))
        modelname += "v";
    Item::set_model(modelname);
}

bool 
LaserBeam::on_laserhit(Direction dir) 
{
    DirectionBits dirbit = to_bits(dir);
    if (!(directions & dirbit)) {
        // `dir' not in `directions' ?
        directions = DirectionBits(directions | dirbit);
        emit_from(get_pos(), dir);
        set_model();
    }
    return false;
}

void 
LaserBeam::on_creation() 
{ 
    if (directions & EASTBIT) emit_from(get_pos(), EAST);
    if (directions & WESTBIT) emit_from(get_pos(), WEST);
    if (directions &NORTHBIT) emit_from(get_pos(), NORTH);
    if (directions &SOUTHBIT) emit_from(get_pos(), SOUTH);
    set_model(); 
}

void
LaserBeam::emit_from(GridPos p, Direction d) 
{
    bool may_pass = true;

    p.move(d);
    if (Stone *st = GetStone(p))
        may_pass = st->on_laserhit(d);

    if (may_pass) {
        Item *it = GetItem(p);
        if (!it) {
            LaserBeam *lb = new LaserBeam(d);
            SetItem(p, lb);
            instances.push_back(lb);
        } else
            it->on_laserhit(d);
    }
}

bool 
LaserBeam::actor_hit(Actor *actor)
{
    double r = actor->get_radius();
    V3 p = actor->get_pos();
    GridPos gp = get_pos();

    // distance of actor from center of the grid
    double dx = fabs(p[0] - gp.x - 0.5) - r;
    double dy = fabs(p[1] - gp.y - 0.5) - r;

    if ((directions & (EASTBIT | WESTBIT) && dy<0.1) ||
        (directions & (NORTHBIT | SOUTHBIT)) && dx<0.1)
    {
        SendMessage(actor, "shatter");
    }

    return false; // laser beams can't be picked up
}

void
LaserBeam::kill_all()
{
    while (!instances.empty())
        world::KillItem(instances[0]->get_pos());
}

void 
LaserBeam::dispose() 
{
    instances.erase(std::find(instances.begin(), instances.end(), this));
    delete this;
}


//----------------------------------------
// Laser stone
//----------------------------------------
vector<LaserStone*> LaserStone::instances;

void
LaserStone::reemit_all()
{
    for (unsigned i=0; i<instances.size(); ++i)
        instances[i]->emit_light();
}

void 
LaserStone::notify_onoff(bool on) 
{
    RecalcLight(); 
}

void 
LaserStone::emit_light()
{
    if (is_on())
        LaserBeam::emit_from(get_pos(), get_dir());
}

void 
LaserStone::on_creation() 
{
    if (is_on())
        LaserBeam::emit_from(get_pos(), get_dir());
    set_model();
}

void 
LaserStone::set_model() 
{
    if (is_on())
        Stone::set_model("st-laseron" + to_suffix(get_dir()));
    else
        Stone::set_model("st-laser" + to_suffix(get_dir()));
}

world::Object *
LaserStone::clone() 
{
    LaserStone *st = new LaserStone;
    instances.push_back(st);
    return st;
}

void 
LaserStone::dispose() 
{
    instances.erase(std::find(instances.begin(),
                              instances.end(), this));
    delete this;
}

//----------------------------------------
// MirrorStone
//----------------------------------------
namespace
{
    class MirrorStone 
        : public MovableStone, public LaserEmitter, public PhotoCell
    {
    protected:
        MirrorStone();

        bool is_transparent() const { return int_attrib("transparent"); }
        bool is_movable() const { return int_attrib("movable"); }

        void set_orientation(int o) {set_attrib("orientation", o);}
        int get_orientation() {return int_attrib("orientation");}

        void emit_light(Direction dir) {
            LaserBeam::emit_from(get_pos(), dir);
            outdirs = DirectionBits(outdirs | to_bits(dir));
        }

        void set_model();
    private:
        // LaserEmitter interface
        DirectionBits emission_directions() const {
            return outdirs;
        }

        // PhotoCell interface
        void on_recalc_start() { outdirs = NODIRBIT; }
        void on_recalc_finish() {}

        // Stone interface
        StoneResponse actor_hit(const StoneContact &sc);
        void on_creation();
        void on_removal();

        // Private methods
        void rotate_right();

        // Variables
        DirectionBits outdirs;
    };
}

MirrorStone::MirrorStone() 
{
    set_attrib("transparent", Value());
    set_attrib("movable", Value());
    set_attrib("orientation", Value(1));
}

void 
MirrorStone::set_model() 
{
    string mname = get_kind();
    mname += is_movable() ? "-m" : "-s";
    mname += is_transparent() ? "t" : "o";
    mname += char('0' + get_orientation());
    Stone::set_model(mname);
}

world::StoneResponse 
MirrorStone::actor_hit(const StoneContact &sc) 
{
    if (is_movable())
        MovableStone::actor_hit(sc);
    rotate_right();
    return world::STONE_REBOUND;
}

void 
MirrorStone::on_creation() 
{
    set_model();
    photo_activate();
}

void 
MirrorStone::on_removal() 
{
    photo_deactivate();
    Stone::on_removal();
}

void 
MirrorStone::rotate_right() 
{
    set_orientation(1+(get_orientation() % 4));
    set_model();
    MaybeRecalcLight(get_pos());
    sound::PlaySound("st-mirrorturn");
}


//----------------------------------------
// PlaneMirror
//----------------------------------------

/** \page st-pmirror Plane Mirror Stone

*/
namespace
{
    class PlaneMirror : 
        public CloneStone<PlaneMirror>, public MirrorStone
    {
    public:
        PlaneMirror() : Stone("st-pmirror") {
            set_orientation('/');
        }
    private:
        void set_orientation(char o) {
            char *a = " -\\|/";
            MirrorStone::set_orientation(strchr(a,o)-a);
        }
        char get_orientation() {
            char *a = " -\\|/";
            return a[MirrorStone::get_orientation()];
        }
        bool on_laserhit(Direction dir);
    };
}

bool 
PlaneMirror::on_laserhit(Direction dir)
{
    switch (get_orientation()) {
    case '|':
        if (dir==EAST || dir==WEST)
            emit_light(reverse(dir));
        break;
    case '-': 
        if (dir==NORTH || dir==SOUTH)
            emit_light(reverse(dir));
        break;
    case '/': 
        switch(dir) {
        case EAST: emit_light(NORTH); break;
        case SOUTH: emit_light(WEST); break;
        case NORTH: emit_light(EAST); break;
        case WEST: emit_light(SOUTH); break;
        case NODIR: break;
        }
        break;
    case '\\': 
        switch(dir) {
        case EAST: emit_light(SOUTH); break;
        case SOUTH: emit_light(EAST); break;
        case NORTH: emit_light(WEST); break;
        case WEST: emit_light(NORTH); break;
        case NODIR: break;
        }
        break;
    }
    if (is_transparent())
        emit_light(dir);
    return false;
}

//----------------------------------------
// TriangleMirror
//----------------------------------------

/** \page st-3mirror Triangle Mirror Stone

*/

namespace
{
    class TriangleMirror : 
        public CloneStone<TriangleMirror>, public MirrorStone
    {
    public:
        TriangleMirror() : Stone("st-3mirror") {}
    private:

        Direction get_orientation() 
        {
            Direction a[] = {NODIR, SOUTH, WEST, NORTH, EAST};
            return a[MirrorStone::get_orientation()];
        }
        bool on_laserhit(Direction dir);
    };
}

bool 
TriangleMirror::on_laserhit(Direction dir) 
{
    Direction orient = get_orientation();

    if (dir == orient) 
        emit_light(reverse(dir));
    else if (dir == reverse(orient)) {
        // this is the "complicated" case where the light falls
        // on the tip of the triangle
        switch (dir) {
        case SOUTH: case NORTH: 
            emit_light(EAST); emit_light(WEST); break;
        case WEST: case EAST: 
            emit_light(SOUTH); emit_light(NORTH); break;
        case NODIR: break;
        }
    } else 
        emit_light(orient);
    if (is_transparent())
        emit_light(dir);
    return false;
}

//----------------------------------------------------------------------
// FUNCTIONS
//----------------------------------------------------------------------
namespace
{
    // This flag is true iff all lasers should be recalculated at the
    // end of the next tick.
    bool light_recalc_scheduled = false;
}

void 
world::laser::Init()
{
    Register(new LaserStone);
    Register(new TriangleMirror);
    Register(new PlaneMirror);
}


void 
world::laser::MaybeRecalcLight(GridPos p)
{
    light_recalc_scheduled |= 
        (LightFrom(p, NORTH) || LightFrom(p, SOUTH) || 
         LightFrom(p, WEST) || LightFrom(p, EAST));
}

void 
world::laser::RecalcLight()
{
    light_recalc_scheduled = true;
}

bool
world::laser::LightFrom(GridPos p, Direction dir)
{
    p.move(dir);
    if (LaserEmitter *le = dynamic_cast<LaserEmitter*>(GetStone(p)))
        if (has_dir(le->emission_directions(), reverse(dir)))
            return true;
    if (LaserEmitter *le = dynamic_cast<LaserEmitter*>(GetItem(p)))
        return (has_dir(le->emission_directions(), reverse(dir)));
    return false;
}

void
world::laser::RecalcLightNow()
{
    if (light_recalc_scheduled) {
        PhotoCell::notify_start();
        LaserBeam::kill_all();
        LaserStone::reemit_all();
        PhotoCell::notify_finish();
        light_recalc_scheduled = false;
    }
}
