//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#ifndef WORLD_HH_INCLUDED
#define WORLD_HH_INCLUDED

#include "enigma.hh"
#include "px/math.hh"
#include "tools.hh"

#include <string>
#include <iosfwd>

namespace world
{
    using actors::Actor;
    using enigma::GridPos;
    using enigma::Direction;
    using enigma::DirectionBits;
    using px::V3;

    using std::string;
}

namespace world
{
    /*
     * This class contains the information the physics engine
     * maintains about dynamic objects ("actors").
     */
    class ActorInfo {
    public:
        ActorInfo(V3 p, V3 v)
            : pos(p), oldpos(p), vel(v), oldvel(v), 
	      charge(0), mass(1), radius(1),
	      grabbed(false)
        {}

        V3 pos, oldpos;		// Absolute position
        V3 vel, oldvel;		// Velocity
	double charge;		// Electric charge
	double mass;		// Mass
	double radius;		// Radius of the sphere
        bool grabbed;		// Actor not controlled by the physics engine
    };

    /*
     * Things that may happen when an actor hits a stone.
     */
    enum StoneResponse {
        STONE_PASS,             // let the actor pass
        STONE_REBOUND,          // rebound the actor
        STONE_KILL,             // rebound the actor and kill the stone
    };

    /*
     * All the information that is necessary to handle 
     * collisions between stones and actors.
     */
    struct StoneContact 
    {
        // Constructor.
        StoneContact(Actor *a, Stone *s,
                     const V3 &cp, const V3 &n);
        
        // Variables.
        Actor           *actor;
        Stone           *stone;
        StoneResponse   response;

        V3      contact_point;  // where do the shapes meet? (world coords)
        V3      normal;         // the surface normal at the contact point
        bool    is_collision;   // actor moves towards the stone, not away
    };

    // Which of the four faces of the stone does the actor touch?
    Direction contact_face(const StoneContact &sc);
    DirectionBits contact_faces(const StoneContact &sc);
}

namespace world
{
    class ForceField {
    public:
	virtual ~ForceField() {}
	virtual V3 get_force(Actor *a, V3 x, V3 v, double time)=0;
	virtual void tick(double dtime) {}
    };

    class ConstantForce : public ForceField {
    public:
	ConstantForce(V3 f) : force(f)
	{}
	V3 get_force(Actor *a,V3 x, V3 v, double time) {
	    return force;
	}
    private:
	V3 force;
    };
}

//----------------------------------------------------------------------
// Global variables
//----------------------------------------------------------------------
namespace world
{
    // The global timer for all objects that need to be notified at
    // regular intervals.
    extern tools::Timer g_timer;
}

namespace world
{
    void Init();

    // Create a new, empty world.
    void Create(int w, int h);
    void Load(const string &name);
    void Tick(double dtime);

    // Destroy all objects etc.
    void Shutdown();

    void NameObject(Object *obj, const string &name);
    Object *GetNamedObject(const string &name);
    void UnnameObject(Object *obj);

//----------------------------------------
// Force fields.
//----------------------------------------
    void AddForceField(ForceField *ff);
    void RemoveForceField(ForceField *ff);
    void SetMouseForce(V3 f);

//----------------------------------------
// Actor access
//----------------------------------------
    void SetActor(double x, double y, Actor* a);
    void GrabActor(Actor *a);
    void ReleaseActor(Actor *a);
    const ActorInfo &GetActorInfo(Actor *a);

//----------------------------------------
// Stone access
//----------------------------------------
    void    SetStone(GridPos p, Stone* st);
    Stone   *GetStone(GridPos p);
    Stone   *YieldStone(GridPos p);
    void    KillStone(GridPos p);

    // Move the stone at position `p' in the specified direction `dir'
    // if (a) there is a stone at `p', (b) the stone is movable, and
    // (c) there is no stone at the new position.
    bool MaybeMoveStone(GridPos p, Direction dir);
    void SwapStones(GridPos p, GridPos newp);

//----------------------------------------
// Item access
//----------------------------------------
    void    SetItem(GridPos p, Item* st);
    void    KillItem(GridPos p);
    Item*   GetItem(GridPos p);

    // Remove item from the world and yield it to the caller.  This is
    // used for transferring items to an inventory.  [Note: This is
    // quite different from `getting' an item with GetItem() in that
    // it transfers ownership to the caller!]
    Item*   YieldItem(GridPos p);

//----------------------------------------
// Floor access
//----------------------------------------
    void    SetFloor(GridPos p, Floor* st);
    Floor*  GetFloor(GridPos p);
    void    KillFloor(GridPos p);

}
#endif
